import {
  schemeCategory10,
  zoom,
  select,
  event,
  csv,
  scaleLinear,
  scaleOrdinal,
  extent
} from 'd3';

const {
  Scene,
  Object3D,
  PerspectiveCamera,
  WebGLRenderer,
  CanvasTexture,
  Sprite,
  SpriteMaterial
} = THREE;

// Data mapping
const dataset =
  'https://gist.githubusercontent.com/curran/a08a1080b88344b0c8a7/raw/d546eaee765268bf2f487608c537c05e22e4b221/iris.csv';
const xValue = d => d.petal_length;
const yValue = d => d.petal_width;
const zValue = d => d.sepal_width;
const colorValue = d => d.species;

// Tweakables
const scaleFactor = 2; // HiDPI
const width = window.innerWidth * scaleFactor;
const height = window.innerHeight * scaleFactor;
const circleSize = 0.15;
const circleOpacity = 0.7;
const circleStrokeWidth = 0.5;
const cameraDistance = 2.3;
const colorScheme = schemeCategory10;
const rotationSensitivity = 0.004;
const rotationIncrement = -0.2;

// Generates a circle sprite for the given color.
// Inspired by https://takahirox.github.io/three.js/examples/canvas_particles_sprites.html
export const generateSprite = color => {
  const canvas = document.createElement('canvas');
  const side = 64;
  const radius = side / 2 - 4;
  canvas.width = side;
  canvas.height = side;

  const context = canvas.getContext('2d');

  // Draw a circle on the canvas.
  context.beginPath();
  context.arc(side / 2, side / 2, radius, 0, 2 * Math.PI, false);
  context.globalAlpha = circleOpacity;
  context.fillStyle = color;
  context.fill();
  context.lineWidth = circleStrokeWidth;
  context.strokeStyle = 'black';
  context.stroke();

  return canvas;
};

// Returns a color for a given datum and number of clusters.
const color = (d, numClusters) => colorScheme[d['clusters_' + numClusters]];

// Cache materials for colors.
const materials = {};
const material = color => {
  if (!materials[color]) {
    materials[color] = new SpriteMaterial({
      map: new CanvasTexture(generateSprite(color))
    });
  }
  return materials[color];
};

const scene = new Scene();
const camera = new PerspectiveCamera(75, width / height, 0.1, 1000);

const renderer = new WebGLRenderer({ alpha: true });
renderer.setSize(width, height);

document.body.appendChild(renderer.domElement);

renderer.domElement.style.width = width / scaleFactor + 'px';
renderer.domElement.style.height = height / scaleFactor + 'px';

const root = new Object3D();
scene.add(root);

camera.position.z = cameraDistance;

// Set up rotation interaction.
const rotationInteraction = zoom();

// Disable zoom, only use panning.
rotationInteraction.scaleExtent([1, 1]);

// Limit panning so it can't go beyond the north and south poles.
// Without this, moving the mouse right and left could cause
// undesirable inverse rotation.
rotationInteraction.translateExtent([
  [-Infinity, -Math.PI / 2 / rotationSensitivity],
  [Infinity, height + Math.PI / 2 / rotationSensitivity]
]);

// Update scenegraph root rotation on pan.
rotationInteraction.on('zoom', () => {
  root.rotation.y = event.transform.x * rotationSensitivity;
  root.rotation.x = event.transform.y * rotationSensitivity;
});

// Set up the interaction to respond to the DOM events.
const rotationSelection = select(renderer.domElement).call(rotationInteraction);

// Initialize rotation
rotationInteraction.translateBy(rotationSelection, 0, 0);

// Rotate slowly all the time.
const animate = () => {
  requestAnimationFrame(animate);
  rotationInteraction.translateBy(rotationSelection, rotationIncrement, 0);
  renderer.render(scene, camera);
};
animate();

const addSprite = ({ x, y, z, color }) => {
  const sprite = new Sprite(material(color));
  sprite.position.x = x;
  sprite.position.y = y;
  sprite.position.z = z;
  sprite.scale.set(circleSize, circleSize, 1);
  root.add(sprite);
};

const row = d => {
  d.petalLength = +d.petalLength;
  d.petalWidth = +d.petalWidth;
  d.sepalLength = +d.sepalLength;
  d.sepalWidth = +d.sepalWidth;
  return d;
};

csv(dataset, row).then(data => {
  const xScale = scaleLinear()
    .domain(extent(data, xValue))
    .range([-1, 1]);

  const yScale = scaleLinear()
    .domain(extent(data, yValue))
    .range([-1, 1]);

  const zScale = scaleLinear()
    .domain(extent(data, zValue))
    .range([-1, 1]);

  const colorScale = scaleOrdinal().range(schemeCategory10);

  data
    .map(d => ({
      x: xScale(xValue(d)),
      y: yScale(yValue(d)),
      z: zScale(zValue(d)),
      color: colorScale(colorValue(d))
    }))
    .forEach(addSprite);
});
